/**
   \file ExternalCommunication.cpp

   \brief Utilitrio para comunicao com aplicaes externas.

   Copyright&copy; 1998-2016 Centro de Sensoriamento Remoto / Universidade
   Federal de Minas Gerais - Brazil.
   All Rights Reserved.

   \author Romulo Fernandes Machado
*/


#include "Config.h"
#define _DFF_CPP_LIBRARY_ BaseLib

#include <cstdlib>
#include <exception>
#include <fstream>
#include <sstream>

#include "Exception.h"
#include "ExceptionUtils.h"
#include "ExternalCommunication.h"
#include "FilesystemUtils.h"
#include "Log.h"


namespace DFF {


namespace {


/**
   \brief Retorna o caminho completo para o arquivo da sesso.
   O caminho retornado  o caminho do diretrio temporrio acrescido
   do nome de base passado pelo usurio e o sufixo '.queue'.

   \param baseName Nome de base a ser usado para sesso.
*/
std::string getSessionName( const std::string& baseName ) {
  return DFF::FilesystemUtils::getCompleteTemporaryPath( baseName + ".queue" );
}

/**
   \brief Retorna se o nome de arquivo passado existe. A funo do
   FilesystemUtils no  utilizada pois ela depende de boost::filesystem::path
   e essa verso  portvel.
*/
bool filenameExists( const std::string& filename ) {
  std::ifstream iStream( filename.c_str(), std::ios_base::in );
  return iStream.good();
}


} // namespace ''


ExternalCommunication::ExternalCommunication() 
  : m_sessionCreated( false ) {
}

ExternalCommunication::~ExternalCommunication() {
  destroy();
}

void ExternalCommunication::createSession( const std::string& sessionIdentifier ) {
  if ( m_sessionCreated )
    throwException( DFF::Exception, "A session named \"" << m_sessionIdentifier <<
      "\" already exists." );

  BOOST_ASSERT( m_sessionIdentifier.empty() );
  BOOST_ASSERT( m_sendQueue == nullptr );
  BOOST_ASSERT( m_receiveQueue == nullptr );

  const std::string receiveQueueName( getSessionName( 
    sessionIdentifier + SHARED_QUEUE_SERVER_SUFFIX ) );

  m_receiveQueue = std::make_shared< FileMessageQueue >( receiveQueueName,
    sizeof( Message ), QUEUE_MAX_MESSAGES );

  logDebug( "External communication receive queue created on \"" << 
    receiveQueueName << "\"." );
  
  const std::string sendQueueName( getSessionName( 
    sessionIdentifier + SHARED_QUEUE_CLIENT_SUFFIX ) );

  m_sendQueue = std::make_shared< FileMessageQueue >( sendQueueName,
    sizeof( Message ), QUEUE_MAX_MESSAGES );

  logDebug( "External communication send queue created on \"" << 
    sendQueueName << "\"." );

  m_sessionCreated = true;
  m_sessionIdentifier = sessionIdentifier;
  logInfo( "External communication session \"" << sessionIdentifier << "\" created." );
}

void ExternalCommunication::destroy() {

  // Apenas apaga os objetos se a instncia tiver sido responsvel
  // pela criao das sesses.
  if ( m_sessionCreated ) {

    logInfo( "Cleaning up External Communication session \"" <<
      m_sessionIdentifier << "\"" );

    m_receiveQueue.reset();
    m_sendQueue.reset();
    destroySession( m_sessionIdentifier );

    m_sessionIdentifier.clear();
    m_sessionCreated = false;
  }
}

void ExternalCommunication::openSession( const std::string& sessionIdentifier ) {
  if ( m_sessionCreated )
    throwException( DFF::Exception, "A session named \"" << m_sessionIdentifier <<
      "\" already exists." );

  BOOST_ASSERT( m_sessionIdentifier.empty() );
  BOOST_ASSERT( m_receiveQueue == nullptr );
  BOOST_ASSERT( m_sendQueue == nullptr );

  try {
    m_sessionIdentifier = sessionIdentifier;
    m_receiveQueue = std::make_shared< FileMessageQueue >(
      getSessionName( m_sessionIdentifier + SHARED_QUEUE_CLIENT_SUFFIX ) );

    m_sendQueue = std::make_shared< FileMessageQueue >(
      getSessionName( m_sessionIdentifier + SHARED_QUEUE_SERVER_SUFFIX ) );

  } catch ( std::exception& ex ) {

    m_sessionIdentifier.clear();
    m_receiveQueue.reset();
    m_sendQueue.reset();

    throwException( DFF::Exception, "Failed to open session \"" << 
      sessionIdentifier << "\", reason: " << ex.what() );
  }
}

void ExternalCommunication::sendMessage( const Message& message, 
  Cancellable* cancellable ) {

  if ( m_sendQueue == nullptr || m_receiveQueue == nullptr )
    throwException( DFF::Exception, "Failed to send message, session is closed." );

  try {
    m_sendQueue->write( &message, sizeof( Message ), cancellable );
  } catch ( std::exception& ex ) {
    throwException( DFF::Exception, "Failed to send external message (identifier: " <<
      message.m_identifier << ", number: " << message.m_number <<
      ", handle: " << message.getFilename() << "). Reason: " <<
      ex.what() << "." );
  }
}

DFF::ExternalCommunication::Message ExternalCommunication::receiveMessage( 
  Cancellable* cancellable ) {

  if ( m_sendQueue == nullptr || m_receiveQueue == nullptr )
    throwException( DFF::Exception, "Failed to send message, session is closed." );

  Message receivedMessage;
  try {
    m_receiveQueue->read( &receivedMessage, sizeof( Message ), cancellable );
    return receivedMessage;

  } catch ( std::exception& ex ) {
    throwException( DFF::Exception, "Failed to receive external message: " << 
      ex.what() );
  }
}

void ExternalCommunication::destroySession( const std::string& identifier ) {

  DFF::FileMessageQueue::destroy( getSessionName( identifier + 
    SHARED_QUEUE_CLIENT_SUFFIX ) );

  DFF::FileMessageQueue::destroy( getSessionName( identifier + 
    SHARED_QUEUE_SERVER_SUFFIX ) );
}

bool ExternalCommunication::sessionExists( const std::string& identifier ) {
  const std::string clientIdentifier( getSessionName( identifier + 
    SHARED_QUEUE_CLIENT_SUFFIX ) );
  const std::string serverIdentifier( getSessionName( identifier + 
    SHARED_QUEUE_SERVER_SUFFIX ) );

  return filenameExists( clientIdentifier ) ||
    filenameExists( serverIdentifier );
}

void ExternalCommunication::storeNumbersOnFilename ( size_t numberCount,
  std::function< double ( size_t ) > getNumber, const std::string& filename ) {

  std::ofstream handle( filename.c_str(), std::ios_base::out |
    std::ios_base::trunc | std::ios_base::binary );

  if ( handle.is_open() == false )
    throwException( DFF::Exception, "Failed to create temporary file \"" <<
      filename << "\" to store number vector." );

  if ( numberCount > std::numeric_limits< int >::max() )
    throwException( DFF::Exception, "List is too big to be communicated. " <<
      "Actual size " << numberCount << ", maximum size " << 
      std::numeric_limits< int >::max() );

  const int listCount = static_cast< int >( numberCount );
  handle.write( ( char* ) &listCount, sizeof( int ) );

  for ( size_t i = 0; i < numberCount; ++i ) {
    const double number = getNumber( i );
    handle.write( ( char* ) &number, sizeof( double ) );
  }

  handle.close();
}

void ExternalCommunication::getNumbersFromFilename( 
  std::function< void( double ) > setNumber, const std::string& filename ) {

  std::ifstream handle( filename.c_str(), std::ios_base::in | 
    std::ios_base::binary );

  if ( handle.is_open() == false )
    throwException( DFF::Exception, "Failed to read temporary file \"" <<
      filename << "\" of stored number vector." );

  int numCount = 0;
  handle.read( ( char* ) &numCount, sizeof( int ) );

  for ( int i = 0; i < numCount; ++i ) {
    double inDouble = 0;
    handle.read( ( char* ) &inDouble, sizeof( double ) );
    setNumber( inDouble );
  }
}

// Evita que bibliotecas externas incluam arquivos do Dinamica.
#ifdef _DINAMICA_EGO_

ExternalCommunication& ExternalCommunicationManager::Instance() {
  return Loki::SingletonHolder< ExternalCommunication, Loki::CreateUsingNew,
    Loki::DefaultLifetime, Loki::ClassLevelLockable >::Instance();
}

#endif // _DINAMICA_EGO_


} // namespace DFF
