/**
   \file ExternalCommunication.h

   \brief Utilitrio para comunicao com aplicaes externas.

   Copyright&copy; 1998-2016 Centro de Sensoriamento Remoto / Universidade
   Federal de Minas Gerais - Brazil.
   All Rights Reserved.

   \author Romulo Fernandes Machado
*/


#ifndef _DINAMICA_EXTERNA_COMMUNICATION_H_
#define _DINAMICA_EXTERNA_COMMUNICATION_H_

#include "Config.h"

#include <array>
#include <functional>
#include <memory>
#include <string>
#include <vector>
#include <thread>

#include "Cancellable.h"
#include "FileMessageQueue.h"

// Evita que bibliotecas externas incluam arquivos do Dinamica.
#ifdef _DINAMICA_EGO_
#  include <Loki/Singleton.h>
#  undef _DFF_H_LIBRARY_
#  define _DFF_H_LIBRARY_ BaseLib
#  include "Library.h"
#endif

#define SHARED_QUEUE_SERVER_SUFFIX "_serverQueue"
#define SHARED_QUEUE_CLIENT_SUFFIX "_clientQueue"
#define QUEUE_MAX_MESSAGES 1024
#define MAX_FILENAME_PATH_SIZE 1024


namespace DFF {


/**
   \brief Tipos bsicos de mensagem trocadas em conexes externas.
*/
enum ExternalMessageTypes {

  /**
     \brief Um nico nmero  enviado no prprio corpo da mensagem.
  */
  EMT_NUMBER = 0,

  /**
     \brief Uma lista de nmeros  passada como contedo de um arquivo.
  */
  EMT_NUMBER_LIST,

  /**
     \brief Uma tabela (nmeros)  passada como contedo de um arquivo.
  */
  EMT_LOOKUP_TABLE,

  /**
     \brief Mensagens customizadas devem ser passadas com identificador
     maior que esse nmero.
  */
  EMT_MAX_MESSAGE_TYPES
};

/**
   \brief Gerencia comunicao com aplicaes externas.
   modelos.
*/
class DFF_PUBLIC ExternalCommunication {
public:

  /**
     \brief Mensagem trocada entre processos.

     No caso da integrao com o R (ou outras ferramentas), os arquivos
     que definem as mensagens nos outros projetos devem ter o mesmo formato
     de mensagem do Dinamica.
  */
  struct Message {

    /**
       \brief Indentificador do contedo / propsito da mensagem.
    */
    uint32_t m_identifier;

    /**
       \brief Espao para que seja gravado um nmero.
    */
    double m_number;

    /**
       \brief Espao para que seja gravado um nome de arquivo.
       O tamanho mximo desse nome de arquivo deve ser MAX_FILENAME_PATH_SIZE.
    */
    std::array< char, MAX_FILENAME_PATH_SIZE > m_filename;

    /**
       \brief Constri uma mensagem vazia.
    */
    Message() 
      : m_identifier( 0 ), 
        m_number( 0 ) {
    
      m_filename.fill( 0 );
    }

    /**
       \brief Constri uma mensagem apenas com identificador.
    */
    explicit Message( uint32_t identifier ) 
      : m_identifier( identifier ),
        m_number( 0 ) {

      m_filename.fill( 0 );
    }

    /**
       \brief Constri uma mensagem com identificador e um nmero.
    */
    explicit Message( uint32_t identifier, double number )
      : m_identifier( identifier ),
        m_number( number ) {

      m_filename.fill( 0 );
    }

    /**
       \brief Constri uma mensagem com identificador e um nome de arquivo.
    */
    explicit Message( uint32_t identifier, 
      const std::array< char, MAX_FILENAME_PATH_SIZE > filename )
      : m_identifier( identifier ),
        m_number( 0 ),
        m_filename( filename ) {
    }

    /**
       \brief Constri uma mensagem com identificador e um nome de arquivo.
    */
    explicit Message( uint32_t identifier, const std::string& filename )
      : m_identifier( identifier ),
        m_number( 0 ) { 

      m_filename.fill( 0 );
      for ( size_t i = 0; i < filename.length(); ++i )
        m_filename[ i ] = filename[ i ];
    }

    /**
       \brief Retorna o nome de arquivo em uma string.
    */
    std::string getFilename() const { 
      std::string filename;

      for ( size_t i = 0; i < m_filename.size(); ++i )
        filename += m_filename[ i ];

      return filename;
    }
  };

  /**
     \brief Constri uma instncia de comunicao com aplicaes externas.

     Essa sesso no troca dados at que seja inicializada ou conectada  
     uma outra sesso j existente.
   */
  ExternalCommunication();

  /**
     \brief Finaliza a sesso interrompendo quaisquer conexes existentes.
   */
  ~ExternalCommunication();

  /**
     \brief Inicializa a sesso para compartilhamento de informaes com aplicaes
     externas.
     (DISPONVEL NO JAVA).
   
     \param sessionIdentifier Identificador da sesso a ser criada.
   
     \exception DFF::Exception Caso a sesso no possa ser criada.
   */
  void createSession( const std::string& sessionIdentifier );

  /**
     \brief Encerra a sesso de compartilhamento.
     (DISPONVEL NO JAVA).
   */
  void destroy();

  /**
     \brief Conecta a uma sesso ativa para compartilhamento de informaes com
     aplicaes externas.

     \param sessionIdentifier Identificador nico de sesso, esse identificador 
     deve respeitar as normas de nome do C++ (impostas pelo boost).
   
     \exception DFF::Exception Caso a sesso no possa ser conectada.
   */
  void openSession( const std::string& sessionIdentifier );

  /**
     \brief Envia uma mensagem atravs da fila de mensagens.

     \param message Mensagem a ser enviada.
     
     \param cancellable Ponteiro para cancellable para verificao de 
     interrupes (opcional).

     Caso a fila esteja cheia, bloqueia at que a mensagem possa ser
     enviada.
  */
  void sendMessage( const Message& message, Cancellable* cancellable = nullptr );

  /**
     \brief Recebe a primeira mensagem da fila.

     \param cancellable Ponteiro para cancellable para verificao de 
     interrupes (opcional).

     A mensagem recebida ser removida da fila e no poder ser recuperada
     novamente. Caso a fila esteja vazia, bloqueia at que uma mensagem seja
     recebida.
  */
  Message receiveMessage( Cancellable* cancellable = nullptr );

  /**
     \brief Utilitrio para remover uma sesso existente pelo nome.

     \param identifier Identificador da sesso a ser destruda.
  */
  static void destroySession( const std::string& identifier );

  /**
     \brief Retorna se uma sesso j existe.
     (DISPONVEL NO JAVA).

     \param identifier Identificador da sesso a ser descoberta.
  */
  static bool sessionExists( const std::string& identifier );

  /**
     \brief Retorna o identificador usado.
     (DISPONVEL NO JAVA).
  */
  inline const std::string& getIdentifier() const {
    return m_sessionIdentifier;
  }

  /**
     \brief Grava uma lista de nmeros no arquivo passado.

     \param numberCount Quantidade de nmeros que devem ser lidos.

     \param getNumber Funo responsvel por retornar cada nmero a partir 
     do ndice pedido.

     \param filename Caminho vlido para o arquivo onde os nmeros sero 
     gravados.
  */
  static void storeNumbersOnFilename( size_t numberCount,
    std::function< double( size_t ) > getNumber,
    const std::string& filename );

  /**
     \brief Recupera uma lista de nmeros do arquivo passado.

     \param setNumber Funo responsvel por receber cada nmero da lista lida.

     \param filename Caminho vlido para o arquivo onde os nmeros esto gravados.
  */
  static void getNumbersFromFilename( std::function< void( double ) > setNumber,
    const std::string& filename );

private:

  /**
     \brief Indica se a sesso aberta foi construda ou apenas conectada
      uma sesso j existente.
  */
  bool m_sessionCreated;

  /**
     \brief Identificador nico da sesso. Esse identificador deve ter as 
     mesmas restries que nomes de variveis em C++ (impostas pelo boost).
  */
  std::string m_sessionIdentifier;

  /**
     \brief Fila de mensagens para envio.
  */
  std::shared_ptr< FileMessageQueue > m_sendQueue;

  /**
     \brief Fila de mensagens para recebimento.
  */
  std::shared_ptr< FileMessageQueue > m_receiveQueue;

}; /* ExternalCommunication */


// Evita que bibliotecas externas incluam arquivos do Dinamica.
#ifdef _DINAMICA_EGO_


/**
   Gerenciador comunicao com aplicaes externas.

   \attention  definido usando herana, e no \c typedef, para que possa
   ser usado fora dos limites dessa biblioteca.
*/
class DFF_PUBLIC ExternalCommunicationManager : public Loki::SingletonHolder< 
  ExternalCommunication, Loki::CreateUsingNew, Loki::DefaultLifetime > {

public:
  static ExternalCommunication& Instance();

}; /* ExternalCommunicationManager */


#endif // _DINAMICA_EGO_


} /* DFF */


#endif /* _DINAMICA_EXTERNA_COMMUNICATION_H_ */
