/**
   \file FileMessageQueue.cpp

   \brief Fila de mensagens utilizando um arquivo para leitura e escrita.

   Copyright&copy; 1998-2016 Centro de Sensoriamento Remoto / Universidade
   Federal de Minas Gerais - Brazil.
   All Rights Reserved.

   \author Romulo Fernandes Machado
*/


#include "Config.h"
#define _DFF_CPP_LIBRARY_ BaseLib

#include <fstream>
#include <chrono>
#include <cstdio>
#include <cstring>
#include <thread>

#include "Exception.h"
#include "ExceptionUtils.h"

#include "FileMessageQueue.h"
#include "FilenameLock.h"
#include "Log.h"

#ifdef _DFF_ON_WINDOWS_
# define _WINDOWS_
#endif

#define BLOCK_SLEEP_INTERVAL_MS 100
#define LOCK_SUFFIX ".lock"


namespace DFF {


namespace {


void writeHeader( std::fstream& stream, 
  const DFF::FileMessageQueue::Header& header ) {

  stream.seekp( 0 );
  stream.write( ( char* ) &header, sizeof( DFF::FileMessageQueue::Header ) );
  stream.flush();
}

template < typename T >
DFF::FileMessageQueue::Header getHeader( T& stream ) {
  DFF::FileMessageQueue::Header header;

  stream.seekg( 0 );
  stream.read( ( char* ) &header, sizeof( DFF::FileMessageQueue::Header ) );

  return header;
}


// Remove um arquivo de forma portvel.
// (Pending) Usar C++ 17 ou bibliotecas do boost para remoo do arquivo.
void removeFile( const std::string& filename ) {
  if ( std::remove( filename.c_str() ) != 0 )
    logDebug( "Failed to remove file \"" << filename << "\"" );
}


}


FileMessageQueue::FileMessageQueue( const std::string& filename, 
  size_t messageSize, size_t messageCount ) 
  : m_filename( filename ),
    m_messageSize( messageSize ) {

  // Obtm a trava exclusiva para criao do arquivo (como supe-se que 
  // o arquivo no existe, dispara se a trava existir).
  const std::string lockFilename( m_filename + LOCK_SUFFIX );
  const FilenameLock lock( lockFilename, FilenameLock::LB_THROW_ON_BUSY );

  // Verifica se a sesso j existe.
  if ( std::ifstream( m_filename, std::ios::in ).is_open() )
    throwException( DFF::Exception, "Failed to create message "
      "queue \"" << m_filename << "\". Session already exist." );

  // Cria o arquivo em modo binrio.
  std::ofstream handle( m_filename, std::ios::out | 
    std::ios::binary | std::ios::trunc );

  if ( handle.is_open() == false )
    throwException( DFF::Exception, "Failed to create message "
      "queue \"" << m_filename << "\"." );

  // Escreve o cabealho com as informaes 
  const FileMessageQueue::Header header = { 0, 0, messageSize, messageCount };
  handle.write( ( char* ) &header, sizeof( FileMessageQueue::Header ) );
  handle.flush();

  // Escreve no arquivo bytes suficientes para acomodar todas as mensagens.
  char* data = new char[ messageSize * messageCount ];
  memset( data, 0, messageSize * messageCount );

  handle.write( data, messageSize * messageCount );
  handle.flush();

  delete [] data;
}

FileMessageQueue::FileMessageQueue( const std::string& filename ) 
  : m_filename( filename ) {

  // Obtm a trava exclusiva para leitura do arquivo.
  const std::string lockFilename( m_filename + LOCK_SUFFIX );
  const FilenameLock lock( lockFilename );

  std::ifstream handle( m_filename, std::ios::in | 
    std::ios::out | std::ios::binary );

  if ( handle.is_open() == false )
    throwException( DFF::Exception, "Failed to open message "
      "queue \"" << m_filename << "\"." );

  // Recupera o cabealho para preencher o tamanho da mensagem armazenada.
  const FileMessageQueue::Header header( getHeader( handle ) );
  m_messageSize = header.blockSize;
}


void FileMessageQueue::write( const void* data, size_t size, 
  Cancellable* cancellable ) {

  while ( true ) {
    const std::string lockFilename( m_filename + ".lock" );
    FilenameLock lock( lockFilename );

    std::fstream handle( m_filename, std::ios::in | 
      std::ios::out | std::ios::binary );

    if ( handle.is_open() == false )
      throwException( DFF::Exception, "Failed to open message "
        "queue \"" << m_filename << "\"." );

    FileMessageQueue::Header header( getHeader( handle ) );

    // Enquanto o ponteiro de escrita no alcanar o ponteiro de leitura,
    // a escrita pode ser feita.
    const size_t nextWritePtr = ( header.writePtr + 1 ) % header.blockCount;
    if ( nextWritePtr != header.readPtr ) {
      size_t offset = sizeof( FileMessageQueue::Header ) + 
        header.writePtr * header.blockSize;

      handle.seekp( offset );
      handle.write( ( const char* ) data, size );
      handle.flush();

      header.writePtr = nextWritePtr;
      writeHeader( handle, header );

      break;
    }

    // Libera o lock e espera um timeout para uma nova tentativa.
    lock.release();
    std::this_thread::sleep_for( std::chrono::milliseconds( BLOCK_SLEEP_INTERVAL_MS ) );

    if ( cancellable != nullptr && cancellable->isCancelled() )
      throwException( DFF::Exception, "Execution aborted." );

  } // while ( true )
}


void FileMessageQueue::read( void* data, size_t size, Cancellable* cancellable ) {
  while ( true ) {
    const std::string lockFilename( m_filename + ".lock" );
    FilenameLock lock( lockFilename );

    std::fstream handle( m_filename, std::ios::in | 
      std::ios::out | std::ios::binary );

    if ( handle.is_open() == false )
      throwException( DFF::Exception, "Failed to open message "
        "queue \"" << m_filename << "\"." );

    FileMessageQueue::Header header( getHeader( handle ) );

    // Uma leitura pode ser feita desde que os dois ponteiros no estejam
    // na mesma posio.
    if ( header.readPtr != header.writePtr ) {
      size_t offset = sizeof( FileMessageQueue::Header ) + 
        header.readPtr * header.blockSize;

      handle.seekp( offset );
      handle.read( ( char* ) data, size );

      header.readPtr = ( header.readPtr + 1 ) % header.blockCount;
      writeHeader( handle, header );

      break;
    }

    // Libera o lock e espera um timeout para uma nova tentativa.
    lock.release();
    std::this_thread::sleep_for( std::chrono::milliseconds( BLOCK_SLEEP_INTERVAL_MS ) );

    if ( cancellable != nullptr && cancellable->isCancelled() )
      throwException( DFF::Exception, "Execution aborted." );

  } // while ( true )
}

void FileMessageQueue::destroy( const std::string& filename ) {

  logDebug( "Destroying file message queue \"" << ( filename + LOCK_SUFFIX ) << "\"" );
  removeFile( filename + LOCK_SUFFIX );

  logDebug( "Destroying file message queue \"" << filename << "\"" );
  removeFile( filename );
}


} // namespace DFF
