/**
   \file FileMessageQueue.h

   \brief Fila de mensagens utilizando um arquivo para leitura e escrita.

   Copyright&copy; 1998-2016 Centro de Sensoriamento Remoto / Universidade
   Federal de Minas Gerais - Brazil.
   All Rights Reserved.

   \author Romulo Fernandes Machado
*/


#ifndef _DFF_FILE_MESSAGE_QUEUE_H_
#define _DFF_FILE_MESSAGE_QUEUE_H_


#include "Config.h"
#include "Cancellable.h"

#include <string>

#ifdef _DINAMICA_EGO_
# undef _DFF_H_LIBRARY_
# define _DFF_H_LIBRARY_ BaseLib
# include "Library.h"
#endif

#ifdef _DFF_ON_WINDOWS_
# define _WINDOWS_
#endif


namespace DFF {


class DFF_PUBLIC FileMessageQueue {
public:

  /**
     \brief Cabealho do arquivo contendo as mensagens.
  */
  struct Header {
    /**
       \brief Posio do cabeote de leitura.
    */
    size_t readPtr;

    /**
       \brief Posio do cabeote de escrita.
    */
    size_t writePtr;

    /**
       \brief Tamanho do bloco de mensagem.
    */
    size_t blockSize;

    /**
       \brief Quantidade de blocos alocados.
    */
    size_t blockCount;
  };

  /**
     \brief Cria um novo arquivo contendo uma fila de mensagens.

     \param filename Nome do arquivo contendo as mensagens.

     \param messageSize Tamanho em bytes das mensagens a serem gravadas.

     \param messageCount Nmero mximo de mensagens a serem gravadas no arquivo.
  */
  FileMessageQueue( const std::string& filename, size_t messageSize, 
    size_t messageCount );

  /**
     \brief Abre um arquivo contendo uma fila de mensagens.

     \param filename Nome do arquivo contendo as mensagens.

     \param accessType O arquivo ser criado para escrita. O nico valor aceito
     para esse parametro  AccessType::OPEN.
  */
  FileMessageQueue( const std::string& filename );

  /**
     \brief Escreve uma mensagem na fila. Bloqueia se a fila estiver cheia.

     \param data Ponteiro para o buffer de dados a ser escrito.

     \param size Nmero de bytes a serem escritos.

     \param cancellable Ponteiro para cancellable para verificao de 
     interrupes (opcional).
  */
  void write( const void* data, size_t size, Cancellable* cancellable = nullptr );

  /**
     \brief L uma mensagem da fila. Bloqueia se a fila estiver vazia.

     \param data Ponteiro para o buffer de dados onde as informaes lidas 
     sero gravadas.

     \param size Nmero de bytes a serem lidos.

     \param cancellable Ponteiro para cancellable para verificao de 
     interrupes (opcional).
  */
  void read( void* data, size_t size, Cancellable* cancellable = nullptr );

  /**
     \brief Destri os arquivos utilizados por uma fila.

     \param filename Nome de arquivo utilizado para criao da fila.
  */
  static void destroy( const std::string& filename );

private:

  /**
     \brief Nome do arquivo onde a fila de mensagem est salva.
  */
  std::string m_filename;

  /**
     \brief Tamanho (em bytes) de cada mensagem armazenada na fila.
  */
  size_t m_messageSize;

}; /* FileMessageQueue */


} /* DFF */


#endif /* _DFF_FILE_MESSAGE_QUEUE_H_ */
