#include <Rcpp.h>

#include <chrono>
#include <cstdio>
#include <fstream>
#include <mutex>
#include <sstream>
#include <thread>

#include "Exception.h"
#include "ExceptionUtils.h"
#include "ExternalCommunication.h"
#include "FilesystemUtils.h"

// General implementation for cancellable (NOT thread safe).
DFF::Cancellable* getCancellable() {
  static DFF::Cancellable lastCancellable;
  return &lastCancellable;
}

// Gerencia a singleton da sessão
DFF::ExternalCommunication& Instance() {
  static DFF::ExternalCommunication* instance = nullptr;
  if ( instance == nullptr )
    instance = new DFF::ExternalCommunication();

  return *instance;
}

//[[Rcpp::export]]
RcppExport SEXP openSession( SEXP identifier ) {
BEGIN_RCPP
  DFF::ExternalCommunication& communicationInstance = Instance();
  communicationInstance.openSession( Rcpp::as< std::string >( identifier ) );

  return Rcpp::wrap( 0 );
END_RCPP
}

//[[Rcpp::export]]
RcppExport SEXP createSession( SEXP identifier ) {
BEGIN_RCPP
  DFF::ExternalCommunication& communicationInstance = Instance();
  communicationInstance.createSession( Rcpp::as< std::string >( identifier ) );

  return Rcpp::wrap( 0 );
END_RCPP
}

//[[Rcpp::export]]
RcppExport SEXP sendNumber( SEXP number ) {
BEGIN_RCPP
  DFF::ExternalCommunication& communicationInstance = Instance();
  communicationInstance.sendMessage( DFF::ExternalCommunication::Message( 
    ( int ) DFF::ExternalMessageTypes::EMT_NUMBER, 
    Rcpp::as< double >( number ) ), getCancellable() );

  return Rcpp::wrap( 0 );
END_RCPP
}

//[[Rcpp::export]]
RcppExport SEXP receiveNumber() {
BEGIN_RCPP
  DFF::ExternalCommunication& communicationInstance = Instance();

  auto message = communicationInstance.receiveMessage( getCancellable() );
  if ( message.m_identifier != DFF::EMT_NUMBER )
    throwException( DFF::Exception, "Received message does not contain a number." );

  return Rcpp::wrap( message.m_number );
END_RCPP
}

//[[Rcpp::export]]
RcppExport SEXP sendNumberVector( SEXP numberVector ) {
BEGIN_RCPP
  Rcpp::NumericVector list( numberVector );
  auto getNumber = [ & ]( size_t index ) -> double {
    return list[ index ];
  };

  const std::string tmpFilename( DFF::FilesystemUtils::getTemporaryFilename() );
  DFF::ExternalCommunication::storeNumbersOnFilename( list.size(), getNumber, tmpFilename );

  try {

    DFF::ExternalCommunication& communicationInstance = Instance();
    communicationInstance.sendMessage( DFF::ExternalCommunication::Message( 
      ( int ) DFF::ExternalMessageTypes::EMT_NUMBER_LIST, 
      tmpFilename ), getCancellable() );

  } catch ( ... ) {

    std::remove( tmpFilename.c_str() );
    throw;
  }

  return Rcpp::wrap( 0 );
END_RCPP
}

//[[Rcpp::export]]
RcppExport SEXP receiveNumberVector() {
BEGIN_RCPP
  DFF::ExternalCommunication& communicationInstance = Instance();

  auto message = communicationInstance.receiveMessage( getCancellable() );
  if ( message.m_identifier != DFF::EMT_NUMBER_LIST )
    throwException( DFF::Exception, "Received message does not contain a number list." );

  std::vector< double > numbers;
  auto setNumber = [ & ]( double number ) {
    numbers.push_back( number );
  };

  DFF::ExternalCommunication::getNumbersFromFilename( setNumber, message.getFilename() );
  std::remove( message.getFilename().c_str() );

  return Rcpp::wrap( numbers );
END_RCPP
}

//[[Rcpp::export]]
RcppExport SEXP sendLookupTable( SEXP keysVector, SEXP valuesVector ) {
BEGIN_RCPP
  Rcpp::NumericVector keys( keysVector );
  Rcpp::NumericVector values( valuesVector );

  if ( keys.size() != values.size() )
    throwException( DFF::Exception, "Keys and values vector must have the same size." );

  const size_t numbersCount = keys.size() + values.size();
  auto getNumber = [ & ]( size_t index ) -> double {
    if ( index < ( size_t ) keys.size() )
      return keys[ index ];
    else
      return values[ index - keys.size() ];
  };

  const std::string tmpFilename( DFF::FilesystemUtils::getTemporaryFilename() );
  DFF::ExternalCommunication::storeNumbersOnFilename( numbersCount, getNumber, tmpFilename );

  try {

    DFF::ExternalCommunication& communicationInstance = Instance();
    communicationInstance.sendMessage( DFF::ExternalCommunication::Message( 
      ( int ) DFF::ExternalMessageTypes::EMT_LOOKUP_TABLE, 
      tmpFilename ), getCancellable() );

  } catch ( ... ) {

    std::remove( tmpFilename.c_str() );
    throw;
  }

  return Rcpp::wrap( 0 );
END_RCPP
}

//[[Rcpp::export]]
RcppExport Rcpp::DataFrame receiveLookupTable() {
BEGIN_RCPP
  DFF::ExternalCommunication& communicationInstance = Instance();

  auto message = communicationInstance.receiveMessage( getCancellable() );
  if ( message.m_identifier != DFF::EMT_LOOKUP_TABLE )
    throwException( DFF::Exception, "Received message does not contain a lookup table." );

  std::vector< double > numbers;
  auto setNumber = [ & ]( double number ) {
    numbers.push_back( number );
  };

  DFF::ExternalCommunication::getNumbersFromFilename( setNumber, message.getFilename() );
  std::remove( message.getFilename().c_str() );

  Rcpp::NumericVector keys; 
  Rcpp::NumericVector values;

  size_t numberOfKeys = numbers.size() / 2;
  for ( size_t i = 0; i < numberOfKeys; ++i ) {
    const double key = numbers[ i ];
    const double value = numbers[ i + numberOfKeys ];

    keys.push_back( key );
    values.push_back( value );
  }

  return Rcpp::DataFrame::create( Rcpp::Named( "Keys" ) = keys,
                                  Rcpp::Named( "Values" ) = values );
END_RCPP
}

